/* ***********************************************************
 * This file was automatically generated on 2019-09-23.      *
 *                                                           *
 * OpenHAB Bindings Version 2.1.23                           *
 *                                                           *
 * If you have a bugfix for this file and want to commit it, *
 * please fix the bug in the generator. You can find a link  *
 * to the generators git repository on tinkerforge.com       *
 *************************************************************/

package com.tinkerforge;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Arrays;
import java.util.List;
import java.net.URI;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.function.Function;
import java.util.function.BiConsumer;
import org.eclipse.smarthome.config.core.Configuration;
import org.eclipse.smarthome.config.core.ConfigDescription;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameter.Type;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameterBuilder;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameterGroup;
import org.eclipse.smarthome.config.core.ParameterOption;
import org.eclipse.smarthome.core.types.State;
import org.eclipse.smarthome.core.types.StateOption;
import org.eclipse.smarthome.core.types.Command;
import org.eclipse.smarthome.core.types.CommandDescriptionBuilder;
import org.eclipse.smarthome.core.types.CommandOption;
import org.eclipse.smarthome.core.thing.ThingTypeUID;
import org.eclipse.smarthome.core.thing.type.ChannelDefinitionBuilder;
import org.eclipse.smarthome.core.thing.type.ChannelType;
import org.eclipse.smarthome.core.thing.type.ChannelTypeBuilder;
import org.eclipse.smarthome.core.thing.type.ChannelTypeUID;
import org.eclipse.smarthome.core.thing.type.ThingType;
import org.eclipse.smarthome.core.thing.type.ThingTypeBuilder;
import org.eclipse.smarthome.core.types.StateDescriptionFragmentBuilder;
import org.eclipse.smarthome.binding.tinkerforge.internal.TinkerforgeBindingConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.eclipse.smarthome.core.library.types.QuantityType;
import org.eclipse.smarthome.core.library.unit.MetricPrefix;
import org.eclipse.smarthome.core.library.unit.SIUnits;
import org.eclipse.smarthome.core.library.unit.SmartHomeUnits;
import org.eclipse.smarthome.core.library.types.OnOffType;

/**
 * Detects inclination of Bricklet (tilt switch open/closed)
 */
public class BrickletTilt extends Device {
	public final static int DEVICE_IDENTIFIER = 239;
	public final static String DEVICE_DISPLAY_NAME = "Tilt Bricklet";

    public final static DeviceInfo DEVICE_INFO = new DeviceInfo(DEVICE_DISPLAY_NAME, "tilt", DEVICE_IDENTIFIER, BrickletTilt.class);

	public final static byte FUNCTION_GET_TILT_STATE = (byte)1;
	public final static byte FUNCTION_ENABLE_TILT_STATE_CALLBACK = (byte)2;
	public final static byte FUNCTION_DISABLE_TILT_STATE_CALLBACK = (byte)3;
	public final static byte FUNCTION_IS_TILT_STATE_CALLBACK_ENABLED = (byte)4;
	public final static byte FUNCTION_GET_IDENTITY = (byte)255;
	private final static int CALLBACK_TILT_STATE = 5;

	public final static short TILT_STATE_CLOSED = (short)0;
	public final static short TILT_STATE_OPEN = (short)1;
	public final static short TILT_STATE_CLOSED_VIBRATING = (short)2;

	private List<TiltStateListener> listenerTiltState = new CopyOnWriteArrayList<TiltStateListener>();

	/**
	 * This listener provides the current tilt state. It is called every time the
	 * state changes.
	 * 
	 * See {@link BrickletTilt#getTiltState()} for a description of the states.
	 */
	public interface TiltStateListener extends DeviceListener {
		public void tiltState(short state);
	}

	/**
	 * Creates an object with the unique device ID \c uid. and adds it to
	 * the IP Connection \c ipcon.
	 */
	public BrickletTilt(String uid, IPConnection ipcon) {
		super(uid, ipcon);

		apiVersion[0] = 2;
		apiVersion[1] = 0;
		apiVersion[2] = 0;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_TILT_STATE)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_ENABLE_TILT_STATE_CALLBACK)] = RESPONSE_EXPECTED_FLAG_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_DISABLE_TILT_STATE_CALLBACK)] = RESPONSE_EXPECTED_FLAG_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_IS_TILT_STATE_CALLBACK_ENABLED)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_IDENTITY)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;

		callbacks[CALLBACK_TILT_STATE] = new IPConnection.DeviceCallbackListener() {
			public void callback(byte[] packet) {
				ByteBuffer bb = ByteBuffer.wrap(packet, 8, packet.length - 8);
				bb.order(ByteOrder.LITTLE_ENDIAN);

				short state = IPConnection.unsignedByte(bb.get());

				for (TiltStateListener listener: listenerTiltState) {
					listener.tiltState(state);
				}
			}
		};
	}

	/**
	 * Returns the current tilt state. The state can either be
	 * 
	 * * 0 = Closed: The ball in the tilt switch closes the circuit.
	 * * 1 = Open: The ball in the tilt switch does not close the circuit.
	 * * 2 = Closed Vibrating: The tilt switch is in motion (rapid change between open and close).
	 * 
	 * .. image:: /Images/Bricklets/bricklet_tilt_mechanics.jpg
	 *    :scale: 100 %
	 *    :alt: Tilt states
	 *    :align: center
	 *    :target: ../../_images/Bricklets/bricklet_tilt_mechanics.jpg
	 */
	public short getTiltState() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_TILT_STATE, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		short state = IPConnection.unsignedByte(bb.get());

		return state;
	}

	/**
	 * Enables the {@link BrickletTilt.TiltStateListener} listener.
	 */
	public void enableTiltStateCallback() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_ENABLE_TILT_STATE_CALLBACK, this);


		sendRequest(bb.array());
	}

	/**
	 * Disables the {@link BrickletTilt.TiltStateListener} listener.
	 */
	public void disableTiltStateCallback() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_DISABLE_TILT_STATE_CALLBACK, this);


		sendRequest(bb.array());
	}

	/**
	 * Returns *true* if the {@link BrickletTilt.TiltStateListener} listener is enabled.
	 */
	public boolean isTiltStateCallbackEnabled() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_IS_TILT_STATE_CALLBACK_ENABLED, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		boolean enabled = (bb.get()) != 0;

		return enabled;
	}

	/**
	 * Returns the UID, the UID where the Bricklet is connected to,
	 * the position, the hardware and firmware version as well as the
	 * device identifier.
	 * 
	 * The position can be 'a', 'b', 'c' or 'd'.
	 * 
	 * The device identifier numbers can be found :ref:`here &lt;device_identifier&gt;`.
	 * |device_identifier_constant|
	 */
	public Identity getIdentity() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_IDENTITY, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		Identity obj = new Identity();
		obj.uid = IPConnection.string(bb, 8);
		obj.connectedUid = IPConnection.string(bb, 8);
		obj.position = (char)(bb.get());
		for (int i = 0; i < 3; i++) {
			obj.hardwareVersion[i] = IPConnection.unsignedByte(bb.get());
		}
		for (int i = 0; i < 3; i++) {
			obj.firmwareVersion[i] = IPConnection.unsignedByte(bb.get());
		}
		obj.deviceIdentifier = IPConnection.unsignedShort(bb.getShort());

		return obj;
	}

	/**
	 * Adds a TiltState listener.
	 */
	public void addTiltStateListener(TiltStateListener listener) {
		listenerTiltState.add(listener);
	}

	/**
	 * Removes a TiltState listener.
	 */
	public void removeTiltStateListener(TiltStateListener listener) {
		listenerTiltState.remove(listener);
	}

    private final Logger logger = LoggerFactory.getLogger(BrickletTilt.class);
    private final static Logger static_logger = LoggerFactory.getLogger(BrickletTilt.class);

    @Override
    public void initialize(org.eclipse.smarthome.config.core.Configuration config, Function<String, org.eclipse.smarthome.config.core.Configuration> getChannelConfigFn, BiConsumer<String, org.eclipse.smarthome.core.types.State> updateStateFn, BiConsumer<String, String> triggerChannelFn) throws TinkerforgeException {
        BrickletTiltConfig cfg = (BrickletTiltConfig) config.as(BrickletTiltConfig.class);
        this.addTiltStateListener((state) -> {if(true) {updateStateFn.accept("TiltTilted", transformTiltTiltedCallback0(state, cfg));}});
		this.addTiltStateListener((state) -> {if(true) {updateStateFn.accept("TiltVibrating", transformTiltVibratingCallback0(state, cfg));}});
        
		{
		BrickletTiltTiltedConfig channelCfg = getChannelConfigFn.apply("TiltTilted").as(BrickletTiltTiltedConfig.class);
		this.enableTiltStateCallback();
		}
		{
		BrickletTiltVibratingConfig channelCfg = getChannelConfigFn.apply("TiltVibrating").as(BrickletTiltVibratingConfig.class);
		this.enableTiltStateCallback();
		}
    }

    @Override
    public void dispose(org.eclipse.smarthome.config.core.Configuration config) throws TinkerforgeException {
        BrickletTiltConfig cfg = (BrickletTiltConfig) config.as(BrickletTiltConfig.class);
        this.listenerTiltState.clear();
		this.listenerTiltState.clear();
        this.disableTiltStateCallback();
		this.disableTiltStateCallback();
		
    }

    @Override
    public List<String> getEnabledChannels(org.eclipse.smarthome.config.core.Configuration config) throws TinkerforgeException{
        BrickletTiltConfig cfg = (BrickletTiltConfig) config.as(BrickletTiltConfig.class);
        List<String> result = new ArrayList<String>();
        result.add("TiltTilted");
		result.add("TiltVibrating");
        return result;
    }

    public static ChannelType getChannelType(ChannelTypeUID channelTypeUID) {
        switch(channelTypeUID.getId()) {
            case "TiltTilted":
                return ChannelTypeBuilder.state(new ChannelTypeUID("tinkerforge", "TiltTilted"), "Tilted", "Switch").withConfigDescriptionURI(URI.create("channel-type:tinkerforge:TiltTilted")).withDescription("The current tilt state. Enabled if tilted, disabled if closed or vibrating.")
.withStateDescription(StateDescriptionFragmentBuilder.create().withReadOnly(true).build().toStateDescription()).build();
            case "TiltVibrating":
                return ChannelTypeBuilder.state(new ChannelTypeUID("tinkerforge", "TiltVibrating"), "Vibrating", "Switch").withConfigDescriptionURI(URI.create("channel-type:tinkerforge:TiltVibrating")).withDescription("The current vibration state. Enabled if vibration is detected, disabled if not. Vibration can only be detected if the bricklet is not tilted.")
.withStateDescription(StateDescriptionFragmentBuilder.create().withReadOnly(true).build().toStateDescription()).build();
            default:
                static_logger.debug("Unknown channel type ID {}", channelTypeUID.getId());
                break;
        }

        return null;
    }

    public static ThingType getThingType(ThingTypeUID thingTypeUID) {
        return ThingTypeBuilder.instance(thingTypeUID, "Tinkerforge Tilt Bricklet").isListed(false).withSupportedBridgeTypeUIDs(Arrays.asList(TinkerforgeBindingConstants.THING_TYPE_BRICK_DAEMON.getId())).withConfigDescriptionURI(URI.create("thing-type:tinkerforge:" + thingTypeUID.getId())).withDescription("Detects inclination of Bricklet (tilt switch open/closed)").withChannelDefinitions(Arrays.asList(new ChannelDefinitionBuilder("TiltTilted", new ChannelTypeUID("tinkerforge", "TiltTilted")).build(), new ChannelDefinitionBuilder("TiltVibrating", new ChannelTypeUID("tinkerforge", "TiltVibrating")).build())).build();
    }

    public static ConfigDescription getConfigDescription(URI uri) {
        switch(uri.toASCIIString()) {
            case "thing-type:tinkerforge:tilt":
                return new ConfigDescription(uri, Arrays.asList(), Arrays.asList(new ConfigDescriptionParameterGroup("update_intervals", "null", true, "Update Intervals", "Update Intervals")));
            case "channel-type:tinkerforge:TiltTilted":
                return new ConfigDescription(uri, Arrays.asList(ConfigDescriptionParameterBuilder.create("updateInterval", Type.INTEGER).withDefault("1000").withDescription("Specifies the update interval in milliseconds. A value of 0 disables automatic updates.").withGroupName("update_intervals").withLabel("Update Interval").withUnit("ms").build()));
            case "channel-type:tinkerforge:TiltVibrating":
                return new ConfigDescription(uri, Arrays.asList(ConfigDescriptionParameterBuilder.create("updateInterval", Type.INTEGER).withDefault("1000").withDescription("Specifies the update interval in milliseconds. A value of 0 disables automatic updates.").withGroupName("update_intervals").withLabel("Update Interval").withUnit("ms").build()));
            default:
                static_logger.debug("Unknown config description URI {}", uri.toASCIIString());
                break;
        }
        return null;
    }

        @Override
    public void refreshValue(String value, org.eclipse.smarthome.config.core.Configuration config, org.eclipse.smarthome.config.core.Configuration channelConfig, BiConsumer<String, org.eclipse.smarthome.core.types.State> updateStateFn, BiConsumer<String, String> triggerChannelFn) throws TinkerforgeException {
        BrickletTiltConfig cfg = (BrickletTiltConfig) config.as(BrickletTiltConfig.class);
        switch(value) {
            case "TiltTilted": {
                   BrickletTiltTiltedConfig channelCfg = channelConfig.as(BrickletTiltTiltedConfig.class);
                   updateStateFn.accept(value, transformTiltTiltedGetter0(this.getTiltState(), cfg));
                   return;
               }
            case "TiltVibrating": {
                   BrickletTiltVibratingConfig channelCfg = channelConfig.as(BrickletTiltVibratingConfig.class);
                   updateStateFn.accept(value, transformTiltVibratingGetter0(this.getTiltState(), cfg));
                   return;
               }
            default:
                logger.warn("Refresh for unknown channel {}", value);
                break;
        }
    }
    

        @Override
    public List<SetterRefresh> handleCommand(org.eclipse.smarthome.config.core.Configuration config, org.eclipse.smarthome.config.core.Configuration channelConfig, String channel, Command command) throws TinkerforgeException {
        List<SetterRefresh> result = Collections.emptyList();
        BrickletTiltConfig cfg = (BrickletTiltConfig) config.as(BrickletTiltConfig.class);
        switch(channel) {
            
            default:
                logger.warn("Command for unknown channel {}", channel);
        }
        return result;
    }

        private org.eclipse.smarthome.core.types.State transformTiltTiltedCallback0(short state, BrickletTiltConfig cfg) {
        return state == 1 ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformTiltVibratingCallback0(short state, BrickletTiltConfig cfg) {
        return state == 2 ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformTiltTiltedGetter0(short value, BrickletTiltConfig cfg) {
        return value == 1 ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformTiltVibratingGetter0(short value, BrickletTiltConfig cfg) {
        return value == 2 ? OnOffType.ON : OnOffType.OFF;
    }
}
